/* -*-mode:C;c-basic-offset:8-*-
 *
 *      Copyright (C) 1993-1995 Bas Laarhoven.
 *
 * Considerable changes by Claus-Justus Heine, 1994, 1995
 * implemented compression and random block access
 *
 * Lot of changes to ftape_flush_buffers and
 * _ftape_write. New function ftape_simple_write()
 *
 *
 

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/ftape-write.c,v $
 $Author: claus $
 *
 $Revision: 1.20 $
 $Date: 1996/03/02 00:57:56 $
 $State: Exp $
 *
 *      This file contains the writing code
 *      for the QIC-117 floppy-tape driver for Linux.
 */

static char RCSid[] =
"$Id: ftape-write.c,v 1.20 1996/03/02 00:57:56 claus Exp $";

  
#include "ftape.h"
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <asm/segment.h>

#include "ftape-write.h"
#include "zftape-buffers.h"
#include "ftape-rw.h"
#include "ecc.h"
#include "ftape-bsm.h"
#include "ftape-ctl.h"
#include "ftape-read.h"
#include "ftape-io.h"
#include "qic117.h"
#include "qic80-compress.h"
#include "qic80-vtbl.h"

/*      Global vars.
 */

/*      Local vars.
 */
int last_write_failed = 0;
int need_flush = 0;

void
ftape_zap_write_buffers( void)
{
	int i;
	
	for (i = 0; i < ftape_num_buffers; ++i) {
		buffer[ i].status = done;
	}
	need_flush = 0;
}

static int
copy_and_gen_ecc( char* destination, byte* source,
                  SectorMap bad_sector_map)
{
	TRACE_FUN( 5, "copy_and_gen_ecc");
	int result;
	struct memory_segment mseg;
	int bads = count_ones( bad_sector_map);
	
	if (bads > 0) {
		TRACEi( 7, "bad sectors in map:", bads);
	}
	if (bads + 3 >= SECTORS_PER_SEGMENT) {
		TRACE( 7, "empty segment");
		mseg.blocks = 0;            /* skip entire segment */
		result = 0;                 /* nothing written */
	} else {
		mseg.blocks = SECTORS_PER_SEGMENT - bads;
		mseg.data = destination;
		memcpy( mseg.data, source, (mseg.blocks - 3) * SECTOR_SIZE);
		result = ecc_set_segment_parity( &mseg);
		if (result < 0) {
			TRACE( 1, "ecc_set_segment_parity failed");
		} else {
			result = (mseg.blocks - 3) * SECTOR_SIZE;
		}
	}
	TRACE_EXIT;
  return result;
}

void
prevent_flush( void)
{
	need_flush = 0;
	ftape_state = idle;
}

int
start_writing( int mode)
{
	TRACE_FUN( 5, "start_writing");
	int result = 0;
	buffer_struct* buff = &buffer[ head];
	int segment_id = buff->segment_id;

	if (ftape_state == writing && buff->status == waiting) {
		setup_new_segment( buff, segment_id, 1);
		if (mode == WRITE_SINGLE) {
			buffer[ head].next_segment = 0; /* stop tape instead 
							 * of pause 
							 */
		}
		calc_next_cluster( buff);     /* prepare */
		buff->status = writing;
		if (runner_status == idle) {
			TRACEi( 7, "starting runner for segment", segment_id);
			result = ftape_start_tape( segment_id, 
						   buff->sector_offset);
			if (result >= 0) {
				runner_status = running;
			}
		}
		if (result >= 0) {
			result = setup_fdc_and_dma( buff, FDC_WRITE); /* go */
		}
		ftape_state = writing;
	}
	TRACE_EXIT;
	return result;
}

int
loop_until_writes_done( void)
{
	TRACE_FUN( 5, "loop_until_writes_done");
	int i;
	int result = 0;

	/*
	 *  Wait until all data is actually written to tape.
	 */
	while (ftape_state == writing && buffer[ head].status != done) {
		TRACEx2( 7, "tail: %d, head: %d", tail, head);
		for (i = 0; i < ftape_num_buffers; ++i) {
			TRACEx3( 8, "buffer[ %d] segment_id: %d, status: %d",
				 i, buffer[ i].segment_id, buffer[ i].status);
		}
		result = fdc_interrupt_wait( 5 * SECOND);
		if (result < 0) {
			TRACE( 1, "fdc_interrupt_wait failed");
			last_write_failed = 1;
			break;
		}
		if (buffer[ head].status == error) {
			/* Allow escape from loop when signaled !
			 */
			if (current->signal & _DONT_BLOCK) {
				TRACE( 2, "interrupted by signal");
				TRACE_EXIT;
				result = -EINTR;          /* is this the right
							   * return value ? 
							   */
				break;
			}
			if (buffer[ head].hard_error_map != 0) {
				/*  Implement hard write error recovery here
				 */
			}
			buffer[ head].status = waiting; /* retry this one */
			if (runner_status == aborting) {
				ftape_dumb_stop();
				runner_status = idle;
			}
			if (runner_status != idle) {
				TRACE( 1, "unexpected state: "
				       "runner_status != idle");
				result = -EIO;
				break;
			}
			start_writing( WRITE_MULTI);
		}
		TRACE( 7, "looping until writes done");
		result = 0;                 /* normal exit status */
	}
	TRACE_EXIT;
	return result;
}

/*      Write given segment from buffer at address onto tape.
 */
int
write_segment( unsigned segment_id, byte* address, int flushing)
{
	TRACE_FUN( 5, "write_segment");
	int result = 0;
	int bytes_written = 0;
  
	TRACEi( 7, "segment_id =", segment_id);
	if (ftape_state != writing) {
		if (ftape_state == reading) {
			TRACE( 7, "calling ftape_abort_operation");
			result = ftape_abort_operation();
			if (result < 0) {
				TRACE( 1, "ftape_abort_operation failed");
			}
		}
		ftape_zap_read_buffers();
		ftape_zap_write_buffers();
		ftape_state = writing;
	}
	/*    if all buffers full we'll have to wait...
	 */
	wait_segment( writing);
	if (buffer[ tail].status == error) {
		/*  setup for a retry
		 */
		buffer[ tail].status = waiting;
		bytes_written = -EAGAIN;    /* force retry */
		if (buffer[ tail].hard_error_map != 0) {
			TRACEx1( 2, "%d hard error(s) in written segment",
				 count_ones( buffer[ tail].hard_error_map));
			TRACEx1( 4, "hard_error_map = 0x%08x", 
				 (unsigned int)buffer[ tail].hard_error_map);
			/*  Implement hard write error recovery here
			 */
		}
	} else if (buffer[ tail].status == done) {
		history.defects += count_ones( buffer[ tail].hard_error_map);
	} else {
		TRACE( 1, "wait for empty segment failed");
		result = -EIO;
	}
	/*    If just passed last segment on tape: wait for BOT or EOT mark.
	 */
	if (result >= 0 && runner_status == logical_eot) {
		int status;

		result = ftape_ready_wait( timeout.seek, &status);
		if ( result < 0 || 
		     (status & (QIC_STATUS_AT_BOT | QIC_STATUS_AT_EOT)) == 0) {
			TRACE( 1, "eot/bot not reached");
		} else {
			runner_status = end_of_tape;
		}
	}
	/*    should runner stop ?
	 */
	if (result >= 0 &&
	    (runner_status == aborting || 
	     runner_status == buffer_underrun ||
	     runner_status == end_of_tape)) {
		if (runner_status != end_of_tape) {
			result = ftape_dumb_stop();
		}
		if (result >= 0) {
			if (runner_status == aborting) {
				if (buffer[ head].status == writing) {
					buffer[ head].status = done; /*????? */
				}
			}
			runner_status = idle;       /* aborted ? */
		}
	}
	/*  Don't start tape if runner idle and segment empty.
	 */
	if (result >= 0 && 
	    !(runner_status == idle &&
	      get_bad_sector_entry( segment_id) == EMPTY_SEGMENT)) {
		if (buffer[ tail].status == done) {
			/*    now at least one buffer is empty, fill it with 
			 *    our data.
			 *    skip bad sectors and generate ecc.
			 *    copy_and_gen_ecc return nr of bytes written,
			 *    range 0..29 Kb inclusive !
			 */
			result = copy_and_gen_ecc( buffer[ tail].address, 
					    address,
					    get_bad_sector_entry( segment_id));
			if (result >= 0) {
				bytes_written = result;
				buffer[ tail].segment_id = segment_id;
				buffer[ tail].status = waiting;
				next_buffer( &tail);
			}
		}
		/*    Start tape only if all buffers full or flush mode.
		 *    This will give higher probability of streaming.
		 */
		if (result >= 0 && runner_status != running &&
		    ((head == tail && buffer[ tail].status == waiting) || 
		     flushing)) {
			result = start_writing( WRITE_MULTI);
		}
	}
	TRACE_EXIT;
	return (result < 0) ? result : bytes_written;
}

/*  Write as much as fits from buffer to the given segment on tape
 *  and handle retries.
 *  Return the number of bytes written (>= 0), or:
 *      -EIO          write failed
 *      -EINTR        interrupted by signal
 *      -ENOSPC       device full
 */
int
_write_segment( unsigned int segment_id, byte* buffer, int flush)
{
	TRACE_FUN( 5, "_write_segment");
	int retry = 0;
	int result;

	history.used |= 2;
	for (;;) {
		if (segment_id > ftape_last_data_segment && !flush) {
			result = -ENOSPC;         /* tape full */
			break;
		}
		result = write_segment( segment_id, buffer, flush);
		if (result < 0) {
			if (result == -EAGAIN) {
				if (++retry > 100) {
					TRACE( 1, "write failed, >100 retries "
					       "in segment");
					result = -EIO;        /* give up */
					break;
				} else {
					TRACEx1( 2, "write error, retry %d", 
						 retry);
				}
			} else {
				TRACEi( 1, "write_segment failed, error:", 
					-result);
				break;
			}
		} else {                    /* success */
			if (result == 0) {        /* empty segment */
				TRACE( 7, "empty segment, nothing written");
			}
			break;
		}
		/* Allow escape from loop when signaled !
		 */
		if (current->signal & _DONT_BLOCK) {
			TRACE( 2, "interrupted by signal");
			TRACE_EXIT;
			result = -EINTR; /* is this the right return value ? */
			break;
		}
	}
	TRACE_EXIT;
	return result;
}

int
ftape_write_header_segments( byte* buffer)
{
	TRACE_FUN( 5, "ftape_write_header_segments");
	int result = 0;
	int retry = 0;
	int header_1_ok = 0;
	int header_2_ok = 0;
	
	result = ftape_abort_operation();
	if ( result < 0 ) {
		TRACE_EXIT;
		return result;
	}
	ftape_seek_to_bot();    /* prevents extra rewind */
	if (GET4( buffer, 0) != 0xaa55aa55) {
		TRACE( 1, "wrong header signature found, aborting");
		return -EIO;
	} 
	if (bad_sector_map_changed) {
		store_bad_sector_map( buffer);
	}
	if ( ftape_label_changed ) {
		memcpy( buffer + 30, ftape_label, 44 );
		PUT4( buffer, 74, 0);
		TRACEx1( 4, "new label: `%s'", (char*)(buffer + 30));
	}
	PUT2( buffer, 144, 0 ); /* nr of eof-marks */
	if ( zftape_compression_map_changed ) {
		PUT2( buffer, 78, zftape_compression_map_location );
	}

	do {
		if (!header_1_ok) {
			while ( (result = _write_segment( header_segment_1, 
							  buffer, 0)) 
				== -EAGAIN );
			if (result < 0) {
				continue;
			}
			header_1_ok = 1;
		}
	} while (result < 0 && retry++ < 3);
	do {
		if (!header_2_ok) {
			while ( (result = _write_segment( header_segment_2, 
							  buffer, 0)) 
				== -EAGAIN );
			if (result < 0) {
				continue;
			}
			header_2_ok = 1;
		}
	} while (result < 0 && retry++ < 3);
	if ( !header_1_ok && !header_2_ok ) {
#ifndef NO_TRACE_AT_ALL
		if (!header_1_ok) {
			TRACE( 1, "update of first header segment failed");
		}
		if (!header_2_ok) {
			TRACE( 1, "update of second header segment failed");
		}
#endif
		result = -EIO;
	}
	TRACE_EXIT;
	return result;
}


static int
prepare_buffers( byte **header_ptr, byte **vtbl_ptr )
{
	TRACE_FUN( 5, "prepare_buffers");
	int result = 0;
	
	*header_ptr = vmalloc( 2*SIZEOF(deblock_buffer) );
	if ( *header_ptr == NULL ) {
		*header_ptr = deblock_buffer;
		*vtbl_ptr = NULL;
	} else {
		*vtbl_ptr = &(*header_ptr)[SIZEOF(deblock_buffer)];
	}
	TRACE( 5, "Reading header segments ... ");
	ftape_seek_to_bot();  /* prevents extra rewind */
	ftape_deblock_segment = -1;
	result = read_segment(used_header_segment, *header_ptr, 
			      *vtbl_ptr != NULL && 
			      ftape_volume_table_changed &&
			      __qic80_num_of_volumes > 0 );
	if ( result < 0 ) {
		TRACE_EXIT;
		return result;
	}
	if ( ftape_volume_table_changed ) {
		PUT2( *header_ptr, 10, first_data_segment );
		if ( *vtbl_ptr ) {
			/*
			 *   read the vtbl segment now to prevent rewinds
			 */
			unsigned seg; 
			
			if ( __qic80_num_of_volumes > 0 ) {
				for ( seg = used_header_segment + 1; 
				      seg < first_data_segment; seg ++ ) {
					(void)read_segment( seg, *vtbl_ptr, 1);
				}
				result = read_segment( first_data_segment,
						       *vtbl_ptr,0 );
				if ( result < 0 ) {
					*vtbl_ptr = NULL;
					result = 0;
				}
			} else {
				memset(*vtbl_ptr,'\0',SIZEOF(deblock_buffer));
			}
		} 
	}
	TRACE_EXIT;
	return result;
}

static void
release_buffers( byte **header_ptr, byte **vtbl_ptr )
{
	if ( *header_ptr != deblock_buffer ) {
		vfree( *header_ptr );
	}
	*header_ptr = *vtbl_ptr = NULL;
}

static int 
verify_header_segments( byte *header_buffer, byte *vtbl_buffer )
{
	int header_1_res, header_2_res, vtbl_res = 0, cmpr_map_res = 0;
	int result = 0;
	int read_ahead;
	
	if ( header_buffer != deblock_buffer ) { /* start verifying */
		read_ahead = 1;
       		header_1_res = 
			zftape_verify_read_segment(header_segment_1, 
						   deblock_buffer, 
						   header_buffer, 
						   SIZEOF(deblock_buffer),
						   1 /* read_ahead */ );
		read_ahead = (ftape_volume_table_changed 
			      || 
			      (zftape_compression_map_location != 0 
			       && zftape_compression_map_changed ));
		header_2_res = 
			zftape_verify_read_segment(header_segment_2, 
						   deblock_buffer, 
						   header_buffer, 
						   SIZEOF(deblock_buffer),
						   read_ahead );
		if ( ftape_volume_table_changed ) {
			read_ahead = (zftape_compression_map_location != 0 
				      && zftape_compression_map_changed );
			vtbl_res=zftape_verify_read_segment(first_data_segment,
							    deblock_buffer, 
							    vtbl_buffer, 
				      ftape_get_seg_size( first_data_segment),
							    read_ahead );
		}
		if ( zftape_compression_map_location != 0 && 
		     zftape_compression_map_changed ) {
			cmpr_map_res = 
		zftape_verify_compression_map(zftape_compression_map_location,
					      deblock_buffer );
		}
		result = (header_1_res >= 0) && 
			(header_2_res >= 0) &&
			(vtbl_res >= 0) && 
			(cmpr_map_res >= 0) ? 0 : -EIO;
	}
	return result;
}
		
		
int
ftape_update_header_segments( void )
{
	TRACE_FUN( 5, "ftape_update_header_segments");
	int result = 0;
	byte *vtbl_buffer = NULL;
	byte *header_buffer = NULL;
	
	if (read_only) {
		TRACE( 4, "Tape set read-only: no update");
		TRACE_EXIT;
		return 0;    /* exit and fake success */
	} 
	if (ftape_state == writing ) {
		result = loop_until_writes_done();
	}
	if ( result >= 0) {
		result = ftape_abort_operation();
		if ( result < 0 ) {
			TRACE_EXIT;
			return result;
		}
	}
	if ( bad_sector_map_changed || zftape_compression_map_changed || 
	     ftape_volume_table_changed || ftape_label_changed ) {
		TRACE( 5, "reading header segments");
		result = prepare_buffers( &header_buffer, &vtbl_buffer );
		if ( result >= 0 ) {
 			result = ftape_write_header_segments( header_buffer );
		}
		if (result >= 0 && ftape_volume_table_changed ) {
			result = qic80_update_volume_table( first_data_segment,
							    vtbl_buffer );
		}
		if ( result >= 0 && 
		     zftape_compression_map_location != 0 && 
		     zftape_compression_map_changed ) {
			result = 
	zftape_update_compression_segments(zftape_compression_map_location,
					   deblock_buffer );
		}
		if ( result >= 0 ) {
			if ( ftape_state == writing ) {
				if( runner_status == idle ) {
					result = start_writing( WRITE_MULTI );
				}
				if ( result >= 0 ) {
					result = loop_until_writes_done();
					prevent_flush();
				}
				if ( result >= 0 ) {
					result = ftape_abort_operation();
				}
				if ( result < 0 ) {
					TRACE_EXIT;
					return result;
				}
			}
			ftape_seek_to_bot();    /* prevents extra rewind */
			result = verify_header_segments( header_buffer, 
							 vtbl_buffer );
			if ( result >= 0 && ftape_state == reading ) {
				result = ftape_abort_operation();
				if ( result >= 0 ) {
					ftape_seek_to_bot();
				}
			}
		}
		if ( result >= 0 ) {
			bad_sector_map_changed = 
				zftape_compression_map_changed = 
				ftape_volume_table_changed = 
				ftape_label_changed = 0;
		}
		release_buffers( &header_buffer, &vtbl_buffer );
	}
	TRACE_EXIT;
	return result;
}

static int 
read_merge_buffer( int data_remaining, int this_segs_size )
{
	TRACE_FUN( 6, "read_merge_buffer");
#ifndef NO_TRACE_AT_ALL
	int old_tracing;
#endif
	int result = 0;
	
	if ( zftape_use_compression ) {
		/*
		 * impossible to insert a compressed block in the middle of
		 * the tape. We can't do anything but append blocks at the end
		 * of the already compressed area.
		 */
		memset( deblock_buffer+data_remaining, 0,
			this_segs_size - data_remaining );
		TRACE_EXIT;
		return 0;
	} else {
		TRACE( 7, "waiting");
		start_writing( WRITE_MULTI);
		result = loop_until_writes_done();
		if (result < 0) {
			TRACE_EXIT;
			return result;
		}      
#ifndef NO_TRACE_AT_ALL
		old_tracing= tracing;
		tracing= -2;   
#endif
		result= read_segment_fraction( ftape_seg_pos, deblock_buffer,
					       0, data_remaining,
					       this_segs_size-data_remaining);
#ifndef NO_TRACE_AT_ALL
		tracing= old_tracing;
#endif
		if ( result != (this_segs_size - data_remaining ) ) {
			TRACEx1( 7, "read_segment result: %d", result);
			memset( deblock_buffer+data_remaining, 0,
				this_segs_size - data_remaining );
			
		}
		TRACE_EXIT;
		return 0;
	}
}

/*
 *  flush the write buffer to tape and write an eof-marker at the
 *  current position if not in raw mode. 
 *  This function always positions the tape before the eof-marker.
 *  _ftape_close() should then advance to the next segment.
 *
 *  the parameter "finish_volume" describes whether to position before
 *  or after the possibly created file-mark. We always position after the
 *  file-mark when called from ftape_close() and a flush was needed (that is
 *  ftape_write() was the last tape operation before calling ftape_flush)
 *  But we always position before the file-mark when this function get's called
 *  from outside ftape_close()
 */
int
ftape_flush_buffers( int finish_volume )
{
	TRACE_FUN( 5, "ftape_flush_buffers");
	int result;
	int data_remaining;
	int this_segs_size;
	static int active = 0;

	if (active) {
		TRACE( 5, "nested call, abort");
		TRACE_EXIT;
		return 0;
	}
	active = 1;
	TRACEi( 7, "entered, ftape_state =", ftape_state);
	if (ftape_state != writing && !need_flush) {
		active = 0;       
		TRACE( 4, "no need for flush");
		TRACE_EXIT;
		return 0;
	}
	if (last_write_failed) {
		ftape_zap_write_buffers();
		active = 0;
		TRACE_EXIT;
		return write_protected ? -EROFS : -EIO;
	}
	TRACE( 4, "flushing write buffers");
	this_segs_size = ftape_get_seg_size( ftape_seg_pos );
	if ( this_segs_size == ftape_seg_data_pos ) {
		ftape_seg_pos ++;
		ftape_seg_data_pos = 0;
	}
	data_remaining  = ftape_seg_data_pos;
	/*
	 *    If there is any data not written to tape yet, append zero's
	 *    up to the end of the sector (if using compression) or merge
	 *    it with the data existing on the tape
	 *    Then write the segment(s) to tape.
	 */
	if ( data_remaining > 0 ) {
		do {
			this_segs_size = ftape_get_seg_size( ftape_seg_pos );
			if ( (this_segs_size > data_remaining) &&
			     (ftape_deblock_segment != ftape_seg_pos) ) {
				result = read_merge_buffer( data_remaining, 
							    this_segs_size );
			        if ( result < 0 ) {
					TRACE( 1, "flush buffers failed");
					need_flush = 0;
					active = 0;
					TRACE_EXIT;
					return result;
				}
				ftape_deblock_segment= ftape_seg_pos;
			}
			result = _write_segment( ftape_seg_pos, 
						 deblock_buffer, 1);
			if (result != this_segs_size ) {
				if (result != -ENOSPC) {
					last_write_failed = 1;
				}
				ftape_data_pos    -= ftape_seg_data_pos;
				ftape_seg_data_pos = 0;
				active = 0;
				TRACE_EXIT;
				return result;
			}
			TRACEi( 7, "flush, moved out buffer:", result);
			/* need next segment for more data (empty segments?)
			 */
			if ( result < data_remaining ) { 
				if (result > 0) {       
					/* move remainder to buffer beginning 
					 */
					memmove(deblock_buffer, 
						deblock_buffer + result,
						SIZEOF(deblock_buffer)-result);
				}
			} 
			data_remaining -= result;
			++ftape_seg_pos;
		} while (data_remaining > 0);
		TRACEi( 7, "result:", result);
		if (data_remaining == 0) {  /* first byte next segment */
			ftape_seg_data_pos = this_segs_size;
			ftape_seg_pos --;
		} else {                    /* after data previous segment */
			ftape_seg_pos -= 1;
			ftape_seg_data_pos = data_remaining + result;
		}
	} else {
		TRACE( 7, "deblock_buffer empty");
		ftape_seg_pos --;
		ftape_seg_data_pos =  ftape_get_seg_size ( ftape_seg_pos );
		start_writing( WRITE_MULTI);
	}
	qic80_maybe_write_eof( ftape_seg_pos, ftape_uncmpr_pos ); 
	if ( finish_volume ) {
		TRACEx2( 4, "Finishing volume at segment %d, "
			 "byte position %d", 
			 ftape_seg_pos, ftape_seg_data_pos );
		zftape_advance_to_next_segment();
		zftape_reset_cmpr_locals();
		ftape_uncmpr_pos = 0;
	} else {
		just_before_eof = zftape_qic80_mode || 
			          ftape_data_pos == ftape_total_data_amount;
	}
	TRACEx2( 7, "ftape_seg_pos: %d, ftape_seg_data_pos: %d",
		 ftape_seg_pos, ftape_seg_data_pos);
	TRACE( 7, "waiting");
	result = loop_until_writes_done();
	if (result < 0) {
		/*
		 * that's really bad. What to to with ftape_data_pos?
		 */
		TRACE( 1, "flush buffers failed");
	}
	ftape_state= idle;         /* low level driver status */
	last_write_failed = 0;
	need_flush = 0;
	active = 0;
	TRACE_EXIT;
	return result;
}

/*
 * return-value: the number of bytes removed from the user-buffer
 * out: 
 *      int *write_cnt: how much actually has been moved to the deblock_buffer
 *      int *req_len  : MUST NOT BE CHANGED, except at EOT, in which case it
 *                      may be adjusted
 * in : 
 *      char *buff          : the user buffer
 *      int buf_pos_write   : copy of  buf_len_wr
 *      int this_segs_size  : the size in bytes of the actual segment
 *      char *deblock_buffer: deblock_buffer
 *
 */
static int
ftape_simple_write ( int *cnt, int *req_len, const char *buff, 
                     int buf_pos_write, int this_seg_size, 
		     char *deblock_buffer )
{
	TRACE_FUN( 5, "ftape_simple_write");
	int remaining;
	int space_left;
	
	/*
	 *  remaining space in this segment, NOT deblock_buffer
	 */
	space_left = this_seg_size - buf_pos_write;
	/*
	 *  what reamains on this tape
	 */
	remaining = ftape_total_data_amount - (ftape_data_pos + *req_len);
	if ( remaining < 0 ) {
		TRACEi( 7, "clipping req_len from:", *req_len );
		*req_len += remaining;
		TRACEi( 7, "to:", *req_len );
	}                           
	*cnt = *req_len < space_left ? *req_len : space_left;
	/*
	 *  already verified in ftape_write()
	 */
	memcpy_fromfs( deblock_buffer + buf_pos_write, buff, *cnt);
	ftape_uncmpr_pos += *cnt;
	TRACE_EXIT;
	return *cnt;
}

static int
check_write_access( int req_len )
{
	TRACE_FUN( 6, "check_write_access");
	int result = 0;
	
	if ( ftape_offline || !formatted || no_tape) {
		TRACEx1( 1,"tape is %s", ftape_offline ?
			 "offline" 
			 : (!formatted ?
			    "not formatted"
			    : "not present"   ));
		result = -EIO;
	} else if ( write_protected ) {
		TRACE( 1, "cartridge write protected");
		last_write_failed = 1;
		result = -EROFS;
	} else if ( first_data_segment == -1) {
		/*
		 *    If we haven't read the header segment yet, do it now.
		 *    This will verify the configuration, get the eof markers
		 *    and the bad sector table and the compression-map.
		 *    We'll use the deblock buffer for scratch.
		 */
		result = read_header_segment( deblock_buffer, 0 );
	}
	if ( result >= 0 ) {
		if ( ftape_old_ftape ) {
			TRACE( 1, "cartridge soft write protected, "
			          "don't write to old ftapes");
			last_write_failed = 1;
			result = -EROFS;
		} else if ( (req_len % ftape_block_size) != 0 ) {
			TRACEx2( 1, "write-count %d must be multiple of "
				    "block-size %d",
				 req_len, ftape_block_size );
			result= -EINVAL;
		} else if( zftape_use_compression ) {
			if ( ftape_data_pos > 
			     (ftape_total_data_amount - ftape_block_size-8)) {
				/* A cluster consumes in the worst case 
				 * ftape_block_size + 8 bytes
				 */
				TRACE( 1, "tape is full");
				result = -ENOSPC;
			} 
		} else if ( ftape_data_pos >= ftape_total_data_amount ) {
			TRACE( 1, "tape is full");
			result = -ENOSPC;         /* full is full */
		}
	}                
	if ( result >= 0 ) {
		result = req_len;
	}
	TRACE_EXIT;
	return result;
}

static int
prepare_contiguous_write_access( void )
{
	TRACE_FUN( 6, "prepare_contigous_write_access" );
	int result = 0;
	int vol_no;

	TRACEi( 7, "ftape_io_state = ", ftape_io_state);
	/*
	 * we need this for our single-sector-access. 
	 * ftape_state doesn't suffice as it gives the "low-level" state
	 * of the function ftape_read/write_segment()
	 */
	ftape_io_state = io_writing;
	/*
	 * provide for single sector access.
	 */
	TRACE( 9, "first call");
	if ( zftape_qic80_mode ) {
		result = qic80_maybe_write_beof( ftape_seg_pos, 
						 just_before_eof != 0 );
		if ( result < 0 ) {
			TRACEi( 1,"qic80_maybe_write_beof() failed with",
				result );
			TRACE_EXIT;
			return result;
		}
		vol_no = qic80_find_file_no( ftape_seg_pos );
		if ( result == 0 && 
		     ftape_block_size != 
		     qic80_get_volume_block_size( vol_no ) ) {
			/*
			 * might be the case if the user does a MTFSFM or 
			 * MTBSFM to append data to an existing file. If the 
			 * user forget about the blocksize he used when 
			 * creating the file, he looses.
			 *
			 *  This caveat is the reason for creating the 
			 *  MTIOC_ZFTAPE_BLKSZ ioctl
			 */
			TRACE( 1, "Can't append to a file with "
			          "different blocksize");
			TRACE_EXIT;
			return -EINVAL;
		}
		TRACEi( 7, "just_before_eof = ", just_before_eof != 0 );
		just_before_eof= 0;
		zftape_compression_map_changed = zftape_use_compression;
	}
	/*
	 *  now merge with old data if neccessary
	 */
	if (ftape_seg_pos != ftape_deblock_segment && ftape_seg_data_pos != 0){
		result = read_segment( ftape_seg_pos, deblock_buffer, 0 );
		ftape_deblock_segment= ftape_seg_pos;
		if (result < 0 && (result != -EINTR) ) {
			TRACEx1( 5, "read_segment result: %d", result);
			TRACE( 5, "Normal if writing to newly formatted tape");
			memset( deblock_buffer, '\0', ftape_seg_data_pos );
			result= 0;
		}
	}
	TRACE_EXIT;
	return result;
}


static int 
write_deblock_buffer_to_tape( int this_seg_size )
{
	TRACE_FUN( 6, "write_buffer_to_tape");
	int result = 0;

	do {
		/* Allow us to escape from this loop with a signal !
		 */
		if (current->signal & _DONT_BLOCK) {
			TRACE_EXIT;
			return -EINTR; /* is this the right return value ? */
		}
		result = _write_segment( ftape_seg_pos, deblock_buffer, 0);
		TRACEi( 7, "_write_segment result =", result);
		if ( result == -EAGAIN ) {
			TRACE( 3, "retry ...");
			continue;
		}
		if ( result != this_seg_size && result ) {
			TRACE( 1, "Uncontinuable error in _write_segment");
			if (result >= 0) {
				TRACEi( 1, "Partial write count ", result);
				result = -EIO;
			}
		} else {
			TRACEi( 7, "moved out of blocking buffer:", result);
			ftape_seg_data_pos =  0;
			ftape_deblock_segment = ftape_seg_pos;
			++ftape_seg_pos;
		}
	} while ( result == -EAGAIN );
	TRACE_EXIT;
	return result;
}


static int
fill_deblock_buffer( int this_seg_size, const char *user_buffer, int *req_len )
{
	TRACE_FUN( 6, "fill_deblock_buffer");
	int cnt;
	int result = 0;

	TRACEi( 7, "remaining req_len:", *req_len);
	TRACEi( 7, "          buf_pos:", ftape_seg_data_pos );
	
	if ( zftape_use_compression ) {
		result= zftape_compress_write( &cnt,
					       req_len,
					       user_buffer,
					       ftape_seg_data_pos,
					       this_seg_size,
					       deblock_buffer );
	} else {
		result= ftape_simple_write( &cnt,
					    req_len, 
					    user_buffer, 
					    ftape_seg_data_pos,
					    this_seg_size, 
					    deblock_buffer );
	}
	/*  deblock_buffer does not contain a valid segment any longer
	 */
	ftape_deblock_segment = -1;
	
	if (result < 0 ) {
		TRACEx2( 1, "ftape_%s_write failed with result %d.",
			 zftape_use_compression ? "compress" : "simple", 
			 result );
		TRACE_EXIT;
		return result;
	}
	TRACEx1( 7, "removed from user-buffer: %05d bytes.", result );
	TRACEx1( 7, "copied to deblock_buffer: %05d bytes.",   cnt );
	
	ftape_seg_data_pos += cnt;
	ftape_data_pos     += cnt;
	TRACEx1( 7, "ftape_data_pos: %d bytes.",  ftape_data_pos );
	TRACE_EXIT;
	return result;
}


/*
 *  called by the kernel-interface routine "ftape_write()"
 */

int
_ftape_write( const char* buff, int req_len)
{
	TRACE_FUN( 5, "_ftape_write");
	int result = 0;
	int written = 0;
	int write_cnt;
	int this_seg_size;
	
	ftape_resid = req_len;
	
	if ( (result = verify_area( VERIFY_READ, buff, req_len ) ) < 0 ) {
		TRACE( 1, "verify_area failed");
		last_write_failed = 1;
		TRACE_EXIT;
		return result;
	}
	/* check if write is allowed 
	 */
	if ( (result = check_write_access( req_len ) ) <= 0 ) { 
		TRACEi( 5, "check_write_access() failed with result", result );
		TRACE_EXIT;
		return result;
	}
	if ( ftape_io_state != io_writing &&
	     (result = prepare_contiguous_write_access() ) < 0 ) {
		TRACEi( 1, "prepare_contigous_write_access() failed with "
			"result", result );
		last_write_failed = 1;
		TRACE_EXIT;
		return result;
	}
	while ( req_len > 0 ) {
		this_seg_size = ftape_get_seg_size( ftape_seg_pos );
		/* Allow us to escape from this loop with a signal !
		 */
		if ( ((result = current->signal&_DONT_BLOCK ? -EINTR:0) < 0) ||
		     ((result= write_cnt= fill_deblock_buffer(this_seg_size, 
							      buff, 
							      &req_len)) < 0)||
		     (ftape_seg_data_pos == this_seg_size &&
		      (result=write_deblock_buffer_to_tape(this_seg_size))<0)){
			TRACEi( 1, "write loop failed with result", result );
			last_write_failed  = 1;
			ftape_resid       -= written;
			TRACE_EXIT;
			return result;
		}
		written += write_cnt;
		buff    += write_cnt;
		req_len -= write_cnt;
	} /* while ( req_len > 0 ) */
	TRACEi( 7, "remaining in blocking buffer:", ftape_seg_data_pos);
	TRACEi( 7, "just written bytes:", written );
	last_write_failed = 0;
	ftape_resid -= written;
	need_flush = written > 0;
	TRACE_EXIT;         
	return written;               /* bytes written */
}
